/********************************************************************************
	Copyright 2005 Sjaak Priester	

	This is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This software is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this software; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

	http://www.gnu.org/licenses/
********************************************************************************/


#include "StdAfx.h"
#include "QEnvelopeText.h"

//================================================================
// QBlockText.cpp

QBlockText::QBlockText()
: m_ScaleMin(0.90f)
, m_ScaleMax(1.10f)
, m_Justification(justFull)
, m_Hyphen(_T('-'))
{
}

QBlockText::~QBlockText()
{
}

int QBlockText::GetBlockText(GraphicsPath& pathOutput, LPCTSTR str, const RectF& rect) const
{
	TEXTMETRIC tm;

	// Retrieve text metrics
	HDC hDC = ::GetDC(NULL);
	HGDIOBJ hOldFont = ::SelectObject(hDC, m_hFont);
	::GetTextMetrics(hDC, & tm);
	::SelectObject(hDC, hOldFont);
	::ReleaseDC(NULL, hDC);

	int nFontHeight = tm.tmAscent - tm.tmInternalLeading;
	int nLeading = tm.tmDescent + tm.tmInternalLeading + tm.tmExternalLeading;
	int nLineHeight = nFontHeight + nLeading;

	// Calculate number of text lines.
	int nLines = ((int) rect.Height + nLeading) / nLineHeight;

	REAL y = rect.Y;

	Strips strips;		// Container for text strips.

	for (int i = 0; i < nLines; ++i)
	{
		strips.push_back(Strip(rect.X, y + nFontHeight - (REAL) tm.tmInternalLeading, rect.Width));
		y += nLineHeight;
	}

	return OutputText(pathOutput, str, strips);
}

bool QBlockText::SetScaleMargins(REAL min, REAL max)
{
	if (min > 0 && min <= 1 && max >= 1)
	{
		m_ScaleMin = min;
		m_ScaleMax = max;
		return true;
	}
	return false;
}

void QBlockText::GetScaleMargins(REAL& min, REAL& max) const
{
	min = m_ScaleMin;
	max = m_ScaleMax;
}

int QBlockText::OutputText(GraphicsPath& pathOutput, LPCTSTR str, const Strips& strips) const
{
	// Distribute the text string str among the strips.
	// I must confess that this is quite an extensive function, and perhaps too long,
	// but splitting it in several functions makes it even more complicated.

	HDC hDC = ::GetDC(NULL);
	HGDIOBJ hOldFont = ::SelectObject(hDC, m_hFont);

	// Collect character distances from string start.
	int nChars = (int) _tcslen(str);
	int * pDx = new int[nChars];
	SIZE sz;
	::GetTextExtentExPoint(hDC, str, nChars, 0, NULL, pDx, &sz);

	// Find the width of the hyphen character.
	::GetTextExtentPoint(hDC, & m_Hyphen, 1, &sz);
	int wHyphen = sz.cx;

	int iBegin = 0;
	int iBreak = 0;

	while (GetBreakType(str, iBegin, nChars) == brkSpace) ++iBegin;	// Skip leading spaces.
	int xBegin = iBegin > 0 ? pDx[iBegin - 1] : 0;

	// For each horizontal strip, take the optimal number of characters.
	for (Strips::const_iterator it = strips.begin(); it != strips.end() && iBegin < nChars - 1; ++it)
	{
		int xMax = xBegin + (int)(it->m_len + 0.5f);

		int iBreakBefore;			// Break indices.
		int iBreakAfter;
		UINT btBefore = brkNone;	// Break types (brkXxx).
		UINT btAfter = brkNone;

		// Step 1:
		// Search for last break position before length will be exceeded.
		for (iBreak = iBegin + 1; iBreak < nChars; ++iBreak)
		{
			UINT b = GetBreakType(str, iBreak, nChars);
			if (b != brkNone)
			{
				iBreakBefore = iBreak;
				btBefore = b;

				if (b == brkReturn) break;
			}
			if (pDx[iBreak] > xMax) break;
		}
		--iBreak;
		// iBreak now contains the 'hard' break index.

		// Step 2:
		// Search for first break position after length is exceeded.
		for (iBreakAfter = iBreak; iBreakAfter < nChars; ++iBreakAfter)
		{
			UINT b = GetBreakType(str, iBreakAfter, nChars);
			if (b != brkNone)
			{
				btAfter = b;
				break;
			}
		}

		// Step 3:
		// Determine optimal break position.
		enum
		{
			hard,
			before,
			after
		} BreakPosition(hard);

		if (btBefore != brkNone)
		{
			if ((m_Justification == justFull || m_Justification == justHard)
				&& btBefore != brkReturn
				&& btAfter != brkNone
				&& btAfter != brkReturn)
			{
				// We've got break positions both before and after.
				int dxBefore = pDx[iBreakBefore - 1] - xBegin;
				if ((btBefore & fDash) != 0) dxBefore += wHyphen;
				REAL scaleBefore = it->m_len / (REAL) dxBefore;		// Stretching (>= 1.0f)

				int dxAfter = pDx[iBreakAfter - 1] - xBegin;
				if ((btAfter & fDash) != 0) dxAfter += wHyphen;
				REAL scaleAfter = it->m_len / (REAL) dxAfter;		// Shrinking (<= 1.0f)

				// Choose between iBreakBefore and iBreakAfter, based on the
				// smallest amount of shrinking or stretching needed.
				BreakPosition = ((1.0f / scaleAfter) < scaleBefore) ? after : before;
			}
			else BreakPosition = before;							// We have only a break position before.
		}
		else if (btAfter != brkNone) BreakPosition = after;			// We have only a break position after.

		// Step 4:
		// Find break type and break index, checking against scale margins.
		UINT BreakType = brkHard;		// Default, in case nothing better is found.
		switch (BreakPosition)
		{
		case before:
			{
				int dx = pDx[iBreakBefore - 1] - xBegin;
				if ((btBefore & fDash) != 0) dx += wHyphen;
				REAL scale = it->m_len / (REAL) dx;					// Stretching (>= 1.0f)
				if (m_Justification != justHard || scale <= m_ScaleMax)
				{
					iBreak = iBreakBefore;
					BreakType = btBefore;
				}
			}
			break;
		case after:
			{
				int dx = pDx[iBreakAfter - 1] - xBegin;
				if ((btAfter & fDash) != 0) dx += wHyphen;
				REAL scale = it->m_len / (REAL) dx;					// Shrinking (< 1.0f)
				if (m_Justification != justHard || scale >= m_ScaleMin)
				{
					iBreak = iBreakAfter;
					BreakType = btAfter;
				}
			}
			break;
		case hard:
		default:
			break;
		}

		// Step 5:
		// Make temporary path with text fragment.
		GraphicsPath pathStrip;
		PointF pnt(it->m_x, it->m_y);

		for (int i = iBegin; i < iBreak; ++i) AddGlyph(str[i], str[i + 1], pnt, & pathStrip);
		if ((BreakType & fDash) != 0) AddGlyph(m_Hyphen, 0, pnt, & pathStrip);

		// Step 6:
		// Calculate scaling.
		RectF rcBounds;
		pathStrip.GetBounds(& rcBounds, 0, 0);

		REAL scale = 1;
		if ((m_Justification == justFull || m_Justification == justHard)
			&& (BreakType & fScale) != 0)
		{
			// Calculate the amount of horizontal scaling.
			scale = it->m_len / rcBounds.Width;
			if (scale < m_ScaleMin) scale = m_ScaleMin;
			else if (scale > m_ScaleMax) scale = m_ScaleMax;
		}

		// Step 7:
		// Scale and translate temporary path.
		Matrix mat;

		switch (m_Justification)
		{
		case justRight:
			mat.Translate(- rcBounds.GetRight(), 0, MatrixOrderAppend);		// Translate right to origin (horizontal axis only).
			mat.Scale(scale, 1.0f, MatrixOrderAppend);						// Scale horizontally.
			mat.Translate(it->m_x + it->m_len, 0, MatrixOrderAppend);		// Translate to right border of strip.
			break;
		case justCenter:
			{
				REAL xCenter = rcBounds.X + rcBounds.Width / 2;
				mat.Translate(- xCenter, 0, MatrixOrderAppend);				// Translate center to origin.
				mat.Scale(scale, 1.0f, MatrixOrderAppend);
				mat.Translate(it->m_x + it->m_len / 2, 0, MatrixOrderAppend);	// Translate to center of strip
			}
			break;
//		case justLeft:
//		case justFull:
//		case justHard:
		default:
			mat.Translate(- rcBounds.GetLeft(), 0, MatrixOrderAppend);		// Translate left to origin.
			mat.Scale(scale, 1.0f, MatrixOrderAppend);
			mat.Translate(it->m_x, 0, MatrixOrderAppend);					// Translate to left of strip
			break;
		}

		pathStrip.Transform(& mat);		// Transform.

		// Step 8:
		// Add temporary path to output.
		pathOutput.AddPath(& pathStrip, FALSE);

		while ((GetBreakType(str, iBreak, nChars) & fSkip) != 0) ++iBreak;	// Skip spaces.
		xBegin = pDx[iBreak - 1];
		iBegin = iBreak;
	}

	delete[] pDx;

	::SelectObject(hDC, hOldFont);
	::ReleaseDC(NULL, hDC);

	return iBreak;
}

// virtual overridable
UINT QBlockText::GetBreakType(LPCTSTR str, int index, int len) const
{
	if (str[index] == _T('\r')) return brkReturn;
	if (index > 0 && str[index -1] == _T('-')) return brkNormal;
	return isspace(str[index]) ? brkSpace : brkNone;
}


//=======================================================================
// QEnvelopeText

enum line
{
	lineNone = 0,
	lineTop = 0x0001,
	lineBottom = 0x0002,
	lineBoth = lineTop | lineBottom
};

QEnvelopeText::QEnvelopeText(const GraphicsPath& pathEnvelope, REAL flatness)
: m_pEnvelope(pathEnvelope.Clone())
{
	m_pEnvelope->Flatten(NULL, flatness);
}

QEnvelopeText::~QEnvelopeText()
{
	delete m_pEnvelope;
}

int QEnvelopeText::GetEnvelopeText(GraphicsPath& pathOutput, LPCTSTR str) const
{
	TEXTMETRIC tm;

	// Retrieve text metrics
	HDC hDC = ::GetDC(NULL);
	HGDIOBJ hOldFont = ::SelectObject(hDC, m_hFont);
	::GetTextMetrics(hDC, & tm);
	::SelectObject(hDC, hOldFont);
	::ReleaseDC(NULL, hDC);

	PathData pd;
	m_pEnvelope->GetPathData(& pd);

	int nFontHeight = tm.tmAscent - tm.tmInternalLeading;
	int nLeading = tm.tmDescent + tm.tmInternalLeading + tm.tmExternalLeading;
	int nLineHeight = nFontHeight + nLeading;

	RectF bounds;
	m_pEnvelope->GetBounds(& bounds);

	// Calculate number of text lines.
	int nLines = ((int) bounds.Height + nLeading) / nLineHeight;
	int nHeight = nLines * nLineHeight - nLeading;

	// Calculate y-coordinate of first top line.
	REAL yTop = bounds.GetTop() + (bounds.Height - nHeight) / 2;

	Strips strips;		// Container for text strips.

	for (int i = 0; i < nLines; ++i)
	{
		HorScan xPoints;			// Container for x-coordinates.

		// Collect the x-coordinates of intersections of top line with envelope path...
		GetXPoints(xPoints, yTop, lineTop);

		// ... and ditto for bottom line.
		REAL yBottom = yTop + nFontHeight;
		GetXPoints(xPoints, yBottom, lineBottom);

		REAL x0(0);
		int lineType(lineNone);

		// Scan horizontal from left to right.
		for (HorScan::const_iterator it = xPoints.begin(); it != xPoints.end(); ++it)
		{
			if (lineType == lineBoth)
			{
				// If just before incrementing, we have a top line and a bottom line, 
				// this must be the end of a strip, containing both a top line and a bottom line.
				// Save this strip.
				REAL x1 = it->first;
				strips.push_back(Strip(x0, yBottom - (REAL) tm.tmInternalLeading, x1 - x0));
			}
			lineType ^= it->second;

			if (lineType == lineBoth)
			{
				// If after incrementing, we have a top line and a bottom line, this
				// must be the beginning of a strip. Remember x-coordinate.
				x0 = it->first;
			}
		}

		yTop += nLineHeight;
	}
	return OutputText(pathOutput, str, strips);
}

void QEnvelopeText::GetXPoints(HorScan& xpoints, REAL y, int lineType) const
{
	// Collect intersections of horizontal line through y with envelope path.
	PathData pd;
	m_pEnvelope->GetPathData(& pd);

	PointF pntStart;
	PointF pntCurrent;
	int nLines;

	for (int i = 0; i < pd.Count; ++i)
	{
		if (pd.Types[i] == PathPointTypeStart)
		{
			pntStart = pd.Points[i];
			pntCurrent = pntStart;
			nLines = 0;
			continue;
		}

		GetXPoint(xpoints, y, lineType, pntCurrent, pd.Points[i]);
		pntCurrent = pd.Points[i];
		if ((pd.Types[i] & PathPointTypeCloseSubpath) != 0 && nLines > 0) GetXPoint(xpoints, y, lineType, pntCurrent, pntStart);
		++nLines;
	}
}

void QEnvelopeText::GetXPoint(HorScan& xpoints, REAL y, int lineType, const PointF& p0, const PointF& p1) const
{
	// If it exists, add intersection of horizontal line through y with
	// line segment p0 - p1.
	REAL y0 = p0.Y;
	REAL y1 = p1.Y;

	if ((y0 <= y && y < y1)
		|| (y1 <= y && y < y0))
	{
		REAL x = p0.X;
		REAL dx = p1.X - x;
		REAL dy = y1 - y0;
		x += dx * (y - y0) / dy;

		xpoints.insert(HorScan::value_type(x, lineType));
	}
}
